% Optimize optObjFunc using the parameters specified in parSetFunc
%
% optObjFunc: 	Handle to an AdjointOptimization class
% parSetFunc: 		Handle to a parameter definition function
function [invParset, trueParset, parset] = optimize(optObjFunc, parSetFunc, invPars, parSetOpts, order, iter, lineSearch, saveMovie, movieName)
default_arg('movieName','Movie');
default_arg('saveMovie', false);
default_arg('order', 2);
default_arg('parSetOpts', struct);
default_arg('invPars', {'K'});
default_arg('parSetFunc', @pars.kInversionPars);
default_arg('optObjFunc', @Acoustic1DOpt);
default_arg('iter', 100);
default_arg('lineSearch', 'golden')

[parset, trueParset] = parSetFunc(parSetOpts);

% Run with true parameters and save data
trueOptObj = optObjFunc(trueParset, invPars, order);
trueOptObj.runForward(false);
trueReceiverData = trueOptObj.forwardReceiverRecordings;
dt = trueOptObj.dt;
trueParset = trueOptObj.pars;
clear trueOptObj;

% Create optimization object
optObj = optObjFunc(parset, invPars, order, dt);
optObj.receiverData = trueReceiverData;
tvec = optObj.tvec;
parset = optObj.pars;

% Compute initial misfit
optObj.runForward(false);
M0 = optObj.computeMisfit();

M = zeros(iter,1);

if saveMovie
	movieName = [movieName '.avi'];
	writerObj = VideoWriter(movieName);
	writerObj.FrameRate = 2;
	open(writerObj);
end

% Setup figure
[upPar, upRec, upMis, figure_handle] = optObj.setupPlot(trueParset, parset, M, M0, iter);

% Golden section parameters
tol = 1e-2;

tic
for i = 1:iter
	% Compute gradient
	grad = optObj.computeGradient();
	M(i) = optObj.computeMisfit();

	upMis(M);
	upRec(optObj);
	upPar(optObj);

	title(sprintf('iter: %d', i-1));
	drawnow;

	if saveMovie
		%===== Add frame to movie ===%
        frame = getframe(gcf);
        writeVideo(writerObj,frame);
        %===========================%
    end

	% One golden section iteration
	relnorm = optObj.gradientNorm(grad);
	L = 0.25*1/relnorm;
    % Returning this new optObj is what updates the parameters
    switch lineSearch
        case 'golden'
            optObj = adjopt.goldenSection(optObj, grad, L, tol);
        case 'jon'
            optObj = adjopt.calcStepLengthJon(optObj, grad);
    end
end
toc

optObj.runForward();
M(end) = optObj.computeMisfit();
upMis(M);

if saveMovie
	%== Close movie object ==%
	close(writerObj);
	%========================%
end

% Return estimated parameters
invParset = optObj.pars;

end
