% Adjoint optimization for 1D acoustic, Cartesian

classdef Acoustic1DOpt < adjopt.AdjointOptimization
    
    properties
        
        xlen
        lim
        
        % Source
        fsource
        xs
        sourceDelta
        adjointReceiverRecordings
        
        % Receiver
        asource
        receiver
        xr
        receiverData
        receiverDelta
        forwardReceiverRecordings
        
        errp
        
        forwardDiscr
        adjointDiscr
        pars      % Current inversion parameters
        
        forwardField % Recorded during latest forward simulation
        adjointField % Recorded during latest adjoint simulation
        
        T    % Final time
        dt    % Time step
        tvec % Includes substage points
        
        % Grid points
        x_primal
        x_dual
        
        m
        order
        H
        Ht             % Time quadrature
        dim
        
        inversionPars
        
    end
    
    methods
        
        function obj = Acoustic1DOpt(pars, inversionPars, order, dt)
            default_arg('inversionPars',{'K'});
            default_arg('dt',1.5e-04);
            obj.xlen = pars.xlen;
            obj.lim = pars.lim;
            obj.order = order;
            obj.fsource = pars.fsource;
            obj.xs = obj.fsource.x;
            obj.receiver = pars.receiver;
            obj.xr = obj.receiver.x;
            obj.dt = dt;
            obj.T  = pars.T;
            rho    = pars.rho;
            K      = pars.K;
            
            % Create grid points
            [g_primal, g_dual] = grid.primalDual1D(obj.xlen, obj.lim);
            obj.x_primal = g_primal.points();
            obj.x_dual = g_dual.points();
            
            % Create discretization objects
            obj.forwardDiscr = obj.createFwdAcousticDiscr(pars);
            obj.adjointDiscr = [];
            
            % Create delta functions
            obj.sourceDelta = obj.createDelta(order,obj.forwardDiscr,obj.fsource.x);
            obj.receiverDelta = obj.createDelta(order,obj.forwardDiscr,obj.receiver.x);
            
            % Set instance variables
            obj.forwardReceiverRecordings = struct;
            obj.forwardReceiverRecordings.p = [];
            obj.forwardReceiverRecordings.v = [];
            obj.adjointReceiverRecordings = struct;
            obj.adjointReceiverRecordings.p = [];
            obj.adjointReceiverRecordings.v = [];
            obj.forwardField = struct;
            obj.forwardField.p = [];
            obj.forwardField.pt = [];
            obj.forwardField.v = [];
            obj.forwardField.vt = [];
            obj.Ht = [];
            obj.pars = pars;
            obj.inversionPars = inversionPars;
        end
        
        function setReceiverData(obj,data)
            obj.receiverData = data;
        end
        
        function fwdDiscr = createFwdAcousticDiscr(obj,pars)
            % Create A
            rho = pars.rho; K = pars.K;
            A = obj.createA(rho,K);
            % Create B
            B = {@(x)0*x, @(x)1 + 0*x; ...
                @(x)1 + 0*x, @(x)0*x};
            
            fwdDiscr = acoustic.DiscretizationPointSourceVariableDiscreteData(...
                obj.xlen, obj.order, obj.lim, [], A, B, [], [], obj.fsource.x, pars.fsource.dat.p);
        end
        
        function adjDiscr = createAdjAcousticDiscr(obj,pars)
            % Create A
            rho = pars.rho; K = pars.K;
            A = obj.createA(rho,K);
            % Create B
            B = {@(x)0*x, @(x)-1 + 0*x; ...
                @(x)-1 + 0*x, @(x)0*x};
            
            %TODO: how to get the adjoint source
            adjDiscr = acoustic.DiscretizationPointSourceVariableDiscreteData(...
                obj.xlen, obj.order, obj.lim, [], A, B, [], [], obj.receiver.x, pars.asource.dat.p);
            
        end
        
        function delta = createDelta(obj,order,discr,xpos)
            delta = struct;
            obj.H = struct;
            % Primal grid
            e_primal = discr.diffOp.e_primal;
            obj.H.primal = e_primal'*discr.diffOp.H*e_primal;
            delta.dp = diracDiscr(xpos, obj.x_primal, order, 0, obj.H.primal);
            % Dual grid
            e_dual = discr.diffOp.e_dual;
            obj.H.dual = e_dual'*discr.diffOp.H*e_dual;
            delta.dv = diracDiscr(xpos, obj.x_dual, order, 0, obj.H.dual);
        end
        
        function A = createA(obj,rho,K)
            ooK = 1./K;
            A = {@(x)(interp1(obj.x_primal,ooK,x)) @(x)0*x; @(x)0*x, @(x)(interp1(obj.x_dual,rho,x))};
        end
        
        function runForward(obj, plotFlag, T)
            default_arg('plotFlag', false);
            default_arg('T', obj.T);
            discr = obj.forwardDiscr;
            receiverDelta = obj.receiverDelta;
            [data, waveField] = obj.runSimulation(discr, T, receiverDelta, plotFlag);
            obj.forwardReceiverRecordings = data;
            obj.forwardField = waveField;
        end
        
        function runAdjoint(obj, plotFlag, T)
            default_arg('plotFlag', false);
            default_arg('T', obj.T);
            discr = obj.adjointDiscr;
            sourceDelta = obj.sourceDelta;
            [data,waveField] = obj.runSimulation(discr, T, sourceDelta, plotFlag);
            obj.adjointReceiverRecordings = data;
            obj.adjointField = waveField;
        end
        
        function [data, waveField] = runSimulation(obj, discr, T, receiverDelta, plotFlag)
            if(plotFlag)
                updatePlotFun = discr.setupPlot();
                plotInterval = 50;
            end
            
            % Set up difference operator, delta function and time stepper
            diffOp = discr.diffOp;
            % Get primal grid (velocity)
            e_primal = diffOp.e_primal;
            % Get dual grid operators
            e_dual = diffOp.e_dual;
            
            % Set zero initial data
            p0_fun = @(x) 0*x; v0_fun = @(x) 0*x;
            p0 = grid.evalOn(discr.diffOp.grid_primal, p0_fun);
            v0 = grid.evalOn(discr.diffOp.grid_dual, v0_fun);
            discr.v0 = [p0; v0];
            
            [ts,nt] = discr.getTimestepper(T,obj.dt);
            [~, b, ~, n_stages] = ts.getTableau();
            
            obj.tvec = zeros(nt*n_stages, 1);
            obj.Ht = kron(ones(nt,1),b);
            
            % Allocate memory
            waveField = struct;
            waveField.p  = zeros(size(e_primal,2),nt*n_stages);
            waveField.pt = zeros(size(e_primal,2),nt*n_stages);
            waveField.v  = zeros(size(e_dual,2),nt*n_stages);
            waveField.vt = zeros(size(e_dual,2),nt*n_stages);
            data = struct;
            data.p = zeros(nt*n_stages, 1);
            data.v = zeros(nt*n_stages, 1);
            
            for it = 1:nt
                ts.step();
                [~, ~, U, Tc, Ut] = ts.getV();
                % Get p
                waveField.p (:,(it-1)*n_stages + 1 : it*n_stages)  = e_primal'*U;
                waveField.pt(:,(it-1)*n_stages + 1 : it*n_stages)  = e_primal'*Ut;
                p = e_primal'*U;
                % Get v
                waveField.v (:,(it-1)*n_stages + 1 : it*n_stages)  = e_dual'*U;
                waveField.vt(:,(it-1)*n_stages + 1 : it*n_stages)  = e_dual'*Ut;
                v = e_dual'*U;
                % Apply delta function to get data
                data.p( (it-1)*n_stages + 1: it*n_stages ) = (obj.H.primal*receiverDelta.dp)'*p;
                data.v( (it-1)*n_stages + 1: it*n_stages ) = (obj.H.dual*receiverDelta.dv)'*v;
                obj.tvec( (it-1)*n_stages + 1: it*n_stages ) = Tc;
                % Look at solution
                if(plotFlag && mod(it,plotInterval)==0)
                    updatePlotFun(ts);
                    drawnow;
                end
            end
        end
        
        %obj.pars contains the updated source
        %pars contains the source at the current iteration
        function updateParameters(obj, pars, direction, steplength)
            for i = 1:numel(obj.inversionPars)
                parName = obj.inversionPars{i};
                if(strcmp(parName,'source'))
                    parName = 'fsource';
                    par = getfield(pars,parName);
                    par.dat.p  = par.dat.p + steplength*direction{i}.p;
                    obj.pars = setfield(obj.pars, parName, par);
                else
                    par = getfield(pars,parName);
                    par = par + steplength*direction{i};
                    obj.pars = setfield(obj.pars, parName, par);
                end
            end
        end
        
        function updateForwardDiscr(obj,pars)
            default_arg('pars', obj.pars);
            
            % Create discr
            obj.forwardDiscr = obj.createFwdAcousticDiscr(pars);
        end
        
        function updateAdjointDiscr(obj,pars)
            pars = obj.pars;
            
            % Make sure that the adjoint are the receivers with the correct
            % data
            data = obj.receiverData;
            approx = obj.forwardReceiverRecordings;
            % Particle velocity
            errv = -(approx.v - data.v);
            asrcv = flipud(errv);
            pars.asource.dat.v = asrcv;
            % Pressure
            obj.errp = -(approx.p - data.p);
            asrcp = flipud(obj.errp);
            pars.asource.dat.p = asrcp;
            
            % Create discr
            obj.adjointDiscr = obj.createAdjAcousticDiscr(pars);
        end
        
        function grad  = computeGradient(obj)
            obj.runForward(false);
            obj.updateAdjointDiscr();
            obj.runAdjoint(false);
            grad = obj.gradientFormula();
        end
        
        %FIXME: This is only for absorbing boundaries for now
        function grad = gradientFormula(obj)
            grad = cell(numel(obj.inversionPars));
            
            for i = 1:numel(obj.inversionPars)
                switch obj.inversionPars{i}
                    case 'source'
                        grad{i}.p = flipud(obj.adjointReceiverRecordings.p);
                        grad{i}.v = flipud(obj.adjointReceiverRecordings.v);
                        
                    case 'rho'
                        curK = getfield(obj.pars,'K');
                        currho = getfield(obj.pars,'rho');
                        gradrho = (obj.forwardField.vt.*fliplr(obj.adjointField.v).*diag(obj.H.dual))*obj.Ht*obj.dt;
                        % Take care of endpoints
                        v0  = obj.forwardField.v(1,:); vn  = obj.forwardField.v(end,:);
                        p0  = obj.forwardField.p(1,:); pn  = obj.forwardField.p(end,:);
                        l20 = obj.adjointField.p(1,:); l2n = obj.adjointField.p(end,:);
                        l30 = obj.adjointField.v(1,:); l3n = obj.adjointField.v(end,:);
                        
                        term1l = (1/4 * 1/sqrt(currho(1)*curK(1)) * curK(1) * v0 .* fliplr(l30))*obj.Ht*obj.dt;
                        term1r = (1/4 * 1/sqrt(currho(end)*curK(end)) * curK(end) * vn .* fliplr(l3n))*obj.Ht*obj.dt;
                        term2l = (-1/4 * 1/((currho(1)*curK(1))^(3/2)) * curK(1) * fliplr(l20) .* p0)*obj.Ht*obj.dt;
                        term2r = (-1/4 * 1/((currho(end)*curK(end))^(3/2)) * curK(end) * fliplr(l2n) .* pn)*obj.Ht*obj.dt;
                        
                        gradrho(1) = gradrho(1) + term1l + term2l;
                        gradrho(end) = gradrho(end) + term1r + term2r;
                        
                        grad{i} = gradrho;
                        
                    case 'K'
                        curK = getfield(obj.pars,'K');
                        currho = getfield(obj.pars,'rho');
                        gradk = -obj.H.primal*((1./(curK).^2).*...
                            (obj.forwardField.pt.*fliplr(obj.adjointField.p))*obj.Ht*obj.dt);
                        % Take care of endpoints
                        v0  = obj.forwardField.v(1,:); vn  = obj.forwardField.v(end,:);
                        p0  = obj.forwardField.p(1,:); pn  = obj.forwardField.p(end,:);
                        l20 = obj.adjointField.p(1,:); l2n = obj.adjointField.p(end,:);
                        l30 = obj.adjointField.v(1,:); l3n = obj.adjointField.v(end,:);
                        
                        term1l = (1/4 * 1/sqrt(currho(1)*curK(1)) * currho(1) * v0 .* fliplr(l30))*obj.Ht*obj.dt;
                        term1r = (1/4 * 1/sqrt(currho(end)*curK(end)) * currho(end) * vn .* fliplr(l3n))*obj.Ht*obj.dt;
                        term2l = (-1/4 * 1/((currho(1)*curK(1))^(3/2)) * currho(1) * fliplr(l20) .* p0)*obj.Ht*obj.dt;
                        term2r = (-1/4 * 1/((currho(end)*curK(end))^(3/2)) * currho(end) * fliplr(l2n) .* pn)*obj.Ht*obj.dt;
                        
                        gradk(1)   = gradk(1) + term1l + term2l;
                        gradk(end) = gradk(end) + term1r + term2r;
                        
                        grad{i} = gradk;
                end
            end
        end
        
        function grad = computeGradientFD(obj,deltaG)
            % Calculate the initial misfit
            pars = obj.pars;
            obj.updateForwardDiscr(pars);
            obj.runForward();
            M0 = obj.computeMisfit();
            for i = 1:numel(obj.inversionPars)
                switch obj.inversionPars{i}
                    case{ 'rho', 'K' }
                        parName = obj.inversionPars{i};
                        par = getfield(pars, parName);
                        grad{i} = zeros(size(par));
                        % Loop over all points in the model
                        for j = 1:length(par)
                            par_temp = par;
                            par_temp(j) = par_temp(j) + deltaG;
                            pars = setfield(pars, parName, par_temp);
                            obj.updateForwardDiscr(pars);
                            obj.runForward(false);
                            M = obj.computeMisfit();
                            dM_dg = (M-M0)/deltaG;
                            grad{i}(j) = dM_dg;
                        end
                    case{ 'source' }
                        parName = obj.inversionPars{i};
                        par = getfield(pars, parName);
                        grad{i} = zeros(size(par.dat.p));
                        % Evaluate function handles
                        if isa(par.dat.p, 'function_handle')
                            par.dat.p = par.dat.p(obj.tvec);
                        end
                        if isa(par.dat.v, 'function_handle')
                            par.dat.v = par.dat.v(obj.tvec);
                        end
                        % Loop over all points in the model
                        for j = 1:length(par.dat.p)
                            par_temp = par;
                            par_temp.dat.p(j) = par_temp.dat.p(j) + deltaG;
                            pars = setfield(pars, parName, par_temp);
                            obj.updateForwardDiscr(pars);
                            obj.runForward(false);
                            M = obj.computeMisfit();
                            dM_dg = (M-M0)/deltaG;
                            grad{i}(j) = dM_dg;
                        end
                end
            end
            
        end
        
        function vInt = timeIntegration(obj, v)
        end
        
        function M = computeMisfit(obj)
            % Calculate error (only p data for now)
            obj.errp = obj.forwardReceiverRecordings.p - obj.receiverData.p;
            errv = obj.forwardReceiverRecordings.v - obj.receiverData.v;
            M = 0.5*obj.errp'*(obj.Ht.*obj.errp)*obj.dt;
        end
        
        function relnorm = gradientNorm(obj, grad)
            relnorm = zeros(1, numel(obj.inversionPars) );
            for i = 1:numel(obj.inversionPars)
                switch obj.inversionPars{i}
                    case 'K'
                        relnorm(i) = norm(grad{i}./obj.pars.K, inf);
                    case 'rho'
                        relnorm(i) = norm(grad{i}./obj.pars.rho, inf);
                    case 'source'
                        if(isa(obj.pars.fsource.dat.p, 'function_handle'))
                            obj.pars.fsource.dat.p = obj.pars.fsource.dat.p(obj.tvec);
                        end
                        relnorm(i) = norm(grad{i}.p,inf)/norm(obj.pars.fsource.dat.p,inf);
                end
            end
            relnorm = max(relnorm);
        end
        
        % Plots the parameter parName for the sets in parsets
        % Example:
        % parName = 'source' (or 'rho')
        % parsets = {parset1, parset2, ...}
        function plotParameterComparison(obj, parName, parsets)
            
            switch parName
                case 'source'
                    % Plot
                    for i = 1:numel(parsets)
                        g = parsets{i}.fsource.dat.p;
                        if isa(g, 'function_handle')
                            g = g(obj.tvec);
                        end
                        plot(obj.tvec(1:4:end), g(1:4:end), 'linewidth', 2);
                        hold on;
                    end
                    hold off;
                    ylimpar = ylim;
                    L = ylimpar(2)-ylimpar(1);
                    ylimpar = [ylimpar(1)-0.1*L, ylimpar(2)+0.1*L];
                    ylim(ylimpar);
                    xlabel('t')
                    ylabel('Source time function')
                    legend('True', 'Current estimate', 'Initial guess');
                   
                case 'rho'
                    % Assume that there are two parameter sets and plot the
                    % relative difference
                    par = cell(1,1);
                    for i = 1:numel(parsets)
                        par{i} = getfield(parsets{i}, 'rho');
                        plot(obj.x_dual, par{i}, 'linewidth', 2);
                        hold on;
                    end
                    hold off;
                    ylimpar = ylim;
                    L = ylimpar(2) - ylimpar(1);
                    ylimpar = [ylimpar(1) - 0.1*L, ylimpar(2) + 0.1*L];
                    ylim(ylimpar);
                    xlabel('x');
                    ylabel('Density');
                    legend('True', 'Current estimate', 'Initial guess');
                case 'K'
                    par = cell(1,1);
                    for i = 1:numel(parsets)
                        par{i} = getfield(parsets{i}, 'K');
                        plot(obj.x_primal, par{i}, 'linewidth', 2);
                        hold on;
                    end
                    hold off;
                    ylimpar = ylim;
                    L = ylimpar(2) - ylimpar(1);
                    ylimpar = [ylimpar(1) - 0.1*L, ylimpar(2) + 0.1*L];
                    ylim(ylimpar);
                    xlabel('x');
                    ylabel('Bulk modulus');
                    legend('True', 'Current estimate', 'Initial guess');
            end
        end
        
        function [upPar, upRec, upMis, figure_handle] = setupPlot(obj, trueParset, initialParset, M, M0, iter)
            scrsz = get(0,'ScreenSize');
            figure_handle = figure('Position',[0.1*scrsz(3) 0.05*scrsz(4) 0.65*scrsz(3) 0.8*scrsz(4)]);
            
            parfig = subplot(3,1,1);
            obj.plotParameterComparison(obj.inversionPars{1} ,{trueParset, obj.pars, initialParset});
            subplot(3,1,2);
            [~, model_handle] = obj.plotPReceiverSignals();
            ylimrec = 1.2*ylim;
            L = ylimrec(2)-ylimrec(1);
            ylimrec = [ylimrec(1)-0.1*L, ylimrec(2)+0.1*L];
            ylim(ylimrec);
            
            subplot(3,1,3);
            misfit = semilogy(M/M0, 'linewidth', 2);
            xlabel('Iterations')
            ylabel('Misfit')
            xlim([1, iter]);
            
            function updateParameter(object)
                axes(parfig);
                object.plotParameterComparison(obj.inversionPars{1} ,{trueParset, object.pars, initialParset});
            end
            
            function updateReceiver(object)
                
                approx = object.forwardReceiverRecordings.p;
                model_handle.YData = approx;
            end
            
            function updateMisfit(M)
                misfit.YData = M/M0;
            end
            
            upPar = @updateParameter;
            upRec = @updateReceiver;
            upMis = @updateMisfit;
        end
        
        function plotGradient(obj, grad)
            
            % Plot first gradient component
            parName = obj.inversionPars{1};
            switch parName
                case 'source'
                    
                case {'rho','lambda','mu'}
                    g = obj.forwardDiscr.grid;
                    
                    grad_mat = grid.funcToPlotMatrix(g, grad{1});
                    X = g.matrices();
                    surf(X{1}', X{2}', grad_mat);
                    view(0,90);
                    shading interp;
                    colorbar
                    ylabel(['Gradient with respect to ' parName])
            end
        end
        
        function [data_handle, model_handle] = plotPReceiverSignals(obj)
            
            data = obj.receiverData;
            approx = obj.forwardReceiverRecordings;
            
            data_handle = plot(obj.tvec, data.p, 'linewidth', 2);
            hold on
            model_handle = plot(obj.tvec, approx.p, 'linewidth', 2);
            xlabel('t')
            legend('Data', 'Model')
            hold off;
        end
        
        function [data_handle, model_handle] = plotVReceiverSignals(obj)
            
            data = obj.receiverData;
            approx = obj.forwardReceiverRecordings;
            
            data_handle = plot(obj.tvec, data.v, 'linewidth', 2);
            hold on
            model_handle = plot(obj.tvec, approx.v, 'linewidth', 2);
            xlabel('t')
            legend('Data', 'Model')
            hold off;
        end
    end
end


